'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var jsep = require('../../parser/jsep.cjs');
var builders = require('../ast/builders.cjs');
var dump = require('../dump.cjs');
var allocState = require('../templates/alloc-state.cjs');
var fnParams = require('../templates/fn-params.cjs');
var internalScope = require('../templates/internal-scope.cjs');
var scope = require('../templates/scope.cjs');
var treeMethod = require('../templates/tree-method.cjs');
var treeMethodCall = require('../templates/tree-method-call.cjs');
var commonjs = require('./modules/commonjs.cjs');
var esm = require('./modules/esm.cjs');
var traversalZones = require('./traversal-zones.cjs');

const params = [builders.identifier('input'), builders.identifier('callbacks')];

const NEW_SCOPE_VARIABLE_DECLARATION = builders.variableDeclaration('const', [
  builders.variableDeclarator(scope.default._, builders.newExpression(builders.identifier('Scope'), params)),
]);

/*
import {
 // deps
} from 'nimma/runtime';
// placement: tree
const tree = {};

// placement: program

export default function (input, callbacks) {
  const scope = new Scope(input, callbacks);

  try {
    // placement: body

    scope.traverse(() => {
      // placement: traverse
    });
  } finally {
    scope.destroy();
  }
}
*/

class ESTree {
  #hashes;
  #tree = builders.objectExpression([]);
  #shorthands = builders.objectExpression([]);
  #runtimeDependencies;
  #traverse = [];
  #availableShorthands;
  #states = -1;

  constructor({ hashes, customShorthands }) {
    this.#hashes = hashes;
    this.cacheInfo = {};
    this.body = [];
    this.traversalZones = new traversalZones.default();
    this.#availableShorthands = customShorthands;
    this.#runtimeDependencies = new Map([['Scope', 'Scope']]);
  }

  addRuntimeDependency(specifier) {
    this.#runtimeDependencies.set(specifier, specifier);
  }

  attachCustomShorthand(name) {
    if (
      this.#availableShorthands === null ||
      !(name in this.#availableShorthands)
    ) {
      throw new ReferenceError(`Shorthand '${name}' is not defined`);
    }

    this.#shorthands.properties.push(
      builders.objectMethod(
        'method',
        builders.identifier(name),
        fnParams.statefulFnParams,
        builders.blockStatement([
          builders.returnStatement(jsep.default(this.#availableShorthands[name])),
        ]),
      ),
    );
  }

  getMethodByHash(hash) {
    return this.#tree.properties.find(prop => prop.key.value === hash);
  }

  addTreeMethod(id, block, scope) {
    this.cacheInfo[id.value] = {
      hash: this.#hashes.getHash(id.value),
      scope,
    };

    if (scope === 'stateful-traverse') {
      const state = allocState.default(++this.#states);
      this.body.push(state);
      this.#tree.properties.push(treeMethod.default(id, block, true));
      this.#traverse.push(treeMethodCall.generateStatefulTreeMethodCall(id, state));
      return;
    }

    this.#tree.properties.push(treeMethod.default(id, block, false));

    const call = treeMethodCall.generateTreeMethodCall(id);
    if (scope === 'traverse') {
      this.#traverse.push(call);
    } else {
      this.body.push(call);
    }
  }

  export(format) {
    const traversalZones = this.traversalZones.root;

    const program = builders.program(
      [
        traversalZones,
        this.#tree.properties.length === 0
          ? null
          : builders.variableDeclaration('const', [
              builders.variableDeclarator(internalScope.default.tree, this.#tree),
            ]),
        this.#shorthands.properties.length === 0
          ? null
          : builders.variableDeclaration('const', [
              builders.variableDeclarator(internalScope.default.shorthands, this.#shorthands),
            ]),
        builders.functionDeclaration(
          null,
          params,
          builders.blockStatement(
            [
              NEW_SCOPE_VARIABLE_DECLARATION,
              builders.tryStatement(
                builders.blockStatement(
                  [
                    ...this.body,
                    this.#traverse.length === 0
                      ? null
                      : builders.expressionStatement(
                          builders.callExpression(scope.default.traverse, [
                            builders.arrowFunctionExpression(
                              [],
                              builders.blockStatement(Array.from(this.#traverse)),
                            ),
                            traversalZones === null
                              ? builders.nullLiteral()
                              : traversalZones.declarations[0].id,
                          ]),
                        ),
                  ].filter(Boolean),
                ),
                null,
                builders.blockStatement([
                  builders.expressionStatement(builders.callExpression(scope.default.destroy, [])),
                ]),
              ),
            ].filter(Boolean),
          ),
        ),
      ].filter(Boolean),
    );

    const mod = format === 'esm' ? esm.default : commonjs.default;
    mod(Array.from(this.#runtimeDependencies), program);

    return dump.default(program);
  }
}

exports.default = ESTree;
